#pragma once

#include <vector>
#include <iostream>
#include <fstream>
#include <queue>
#include <algorithm>
#include <set>

const int INF = 1 << 20;

class Graph {
private:
    int const s;
    int const t;
    int const n;
    int const m;
    int const l;
    std::vector<std::pair<int, int>> const edge_list;
    std::vector<std::vector<std::pair<int, int>>> const edge;
    std::vector<std::vector<int>> dist;
    int createDist;

public:
    Graph(int s, int t, int n, int m, int l, std::vector<std::pair<int, int>> edge_list, std::vector<std::vector<std::pair<int, int>>> edge)
        :s(s), t(t), n(n), m(m), l(l), edge_list(edge_list), edge(edge) {
        createDist = false;
    }

    void constructDist() {
        dist.assign(n, std::vector<int>(n, INF));
        for(int i = 0; i < m; i++) {
            int u = edge_list[i].first;
            int v = edge_list[i].second;
            dist[u][v] = 1;
            dist[v][u] = 1;
        }
        for(int i = 0; i < n; i++) dist[i][i] = 0;
        for(int k = 0; k < n; k++) {
            for(int i = 0; i < n; i++) {
                for(int j = 0; j < n; j++) dist[i][j] = std::min(dist[i][j], dist[i][k] + dist[k][j]);
            }
        }
        createDist = true;
    }

    std::pair<int, int> getEdge(int e) const {
        return std::make_pair(edge_list[e].first, edge_list[e].second);
    }

    std::vector<std::pair<int, int>> getEdges() const {
        return edge_list;
    }

    std::vector<std::vector<std::pair<int, int>>> getGraph() const {
        return edge;
    }

    int getStart() const {
        return s;
    }

    int getTerminal() const {
        return t;
    }

    int numVertices() const {
        return n;
    }

    int numEdges() const {
        return m;
    }

    int numLength() const {
        return l;
    }

    int getDistance(int u, int v) {
        if(createDist == false) constructDist();
        return dist[u][v];
    }

    std::vector<std::pair<int, int>> getNeighbors(int v) const {
        return edge[v];
    }

    void printEdges() const {
        for(auto it = edge_list.begin(); it != edge_list.end(); ++it) {
            std::cout << (*it).first << ", " << (*it).second << std::endl;
        }
    }

    void print() const {
        std::cout << "Vertices: " << numVertices() << std::endl;
        std::cout << "Edges: "    << numEdges() << std::endl;
        std::cout << "Length: "   << numLength() << std::endl;
        std::cout << "Terminal: " << getStart() << ", " << getTerminal() << std::endl;
        std::cout << "Edge List: " << std::endl;
        printEdges();
    }

    Graph deleteVertices(int start, int terminal, int length) {
        if(createDist == false) constructDist();
        std::vector<int> validVertices(n, true);
        std::vector<int> numVertices(n);
        for(int k = 0; k < n; k++) {
            if(dist[start][k] + dist[k][terminal] > length) validVertices[k] = false;
        }
        int numValidVertices = 0;
        int numValidEdges = 0;
        int idx = 0;
        for(int i = 0; i < n; i++) {
            if(validVertices[i]) {
                numValidVertices++;
                numVertices[i] = idx++;
            }
        }
        std::vector<std::pair<int, int>> res_edge_list;
        for(int i = 0; i < m; i++) {
            int u = edge_list[i].first;
            int v = edge_list[i].second;
            if(validVertices[u] && validVertices[v]) {
                res_edge_list.push_back(std::make_pair(numVertices[u], numVertices[v]));
                numValidEdges++;
            }
        }
        std::vector<std::vector<std::pair<int, int>>> res_edge(numValidVertices);
        for(auto it = res_edge_list.begin(); it != res_edge_list.end(); ++it) {
            int u = (*it).first;
            int v = (*it).second;
            res_edge[u].push_back(std::make_pair(v, 1));
            res_edge[v].push_back(std::make_pair(u, 1));
        }
        Graph res(numVertices[start], numVertices[terminal], numValidVertices, numValidEdges, length, res_edge_list, res_edge);
        return res;
    }

    Graph deleteLeaves() {
        std::vector<int> validVertices(n, true);
        std::vector<int> numVertices(n);
        std::vector<std::set<int>> graph(n);
        for(int i = 0; i < m; i++) {
            int u = edge_list[i].first;
            int v = edge_list[i].second;
            graph[u].insert(v);
            graph[v].insert(u);
        }
        for(int i = 0; i < n; i++) {
            if(i == s || i == t) continue;
            if(graph[i].size() == 1) {
                int u = i;
                int v = *graph[u].begin();
                deleteLeaf(graph, u, v);
                while(graph[v].size() == 1) {
                    u = v;
                    v = *graph[u].begin();
                    deleteLeaf(graph, u, v);
                }
            }
        }
        for(int i = 0; i < n; i++) {
            if(graph[i].size() == 0) validVertices[i] = false;
        }
        int numValidVertices = 0;
        int numValidEdges = 0;
        int idx = 0;
        for(int i = 0; i < n; i++) {
            if(validVertices[i]) {
                numValidVertices++;
                numVertices[i] = idx++;
            }
        }
        std::vector<std::pair<int, int>> res_edge_list;
        for(int u = 0; u < n; u++) {
            for(auto v: graph[u]) {
                if(u < v) {
                    res_edge_list.push_back(std::make_pair(numVertices[u], numVertices[v]));
                    numValidEdges++;
                }
            }
        }
        std::vector<std::vector<std::pair<int, int>>> res_edge(numValidVertices);
        for(auto it = res_edge_list.begin(); it != res_edge_list.end(); ++it) {
            int u = (*it).first;
            int v = (*it).second;
            res_edge[u].push_back(std::make_pair(v, 1));
            res_edge[v].push_back(std::make_pair(u, 1));
        }
        Graph res(numVertices[s], numVertices[t], numValidVertices, numValidEdges, l, res_edge_list, res_edge);
        return res;
    }

    void deleteLeaf(std::vector<std::set<int>> &graph, int u, int v) {
        graph[u].erase(v);
        graph[v].erase(u);
    }

    Graph duplicate(int start, int terminal, int length) {
        Graph res(start, terminal, n, m, length, edge_list, edge);
        return res;
    }
};

Graph readGraph() {
    char c;
    int n, m, l;
    int s = -1, t = -1;
    std::vector<std::pair<int, int>> edge_list;
    char buf[128];
    while (fgets(buf, sizeof(buf), stdin)) {
        if (buf[0] == 'p') {
            sscanf(buf, "p edge %d %d", &n, &m);
        }
        else if (buf[0] == 'e') {
            int u, v;
            sscanf(buf, "e %d %d", &u, &v);
            u--;
            v--;
            edge_list.push_back(std::make_pair(u, v));
        }
        else if (buf[0] == 'l') {
            sscanf(buf, "l %d", &l);
        }
        else if (buf[0] == 't') {
            sscanf(buf, "t %d %d", &s, &t);
            s--;
            t--;
        }
    }
    std::sort(edge_list.begin(), edge_list.end());
    std::vector<std::vector<std::pair<int, int>>> edge(n);
    for(auto it = edge_list.begin(); it != edge_list.end(); ++it) {
        int u = (*it).first;
        int v = (*it).second;
        edge[u].push_back(std::make_pair(v, 1));
        edge[v].push_back(std::make_pair(u, 1));
    }
    Graph G(s, t, n, m, l, edge_list, edge);
    return G;
}

void writeGraph(Graph G) {
    int n = G.numVertices();
    int m = G.numEdges();
    int l = G.numLength();
    int s = G.getStart()    + 1;
    int t = G.getTerminal() + 1;
    FILE *fp = fopen("outfile.col", "wt");
    fprintf(fp, "p edge %d %d\n", n, m);
    for(int i = 0; i < m; i++) {
        int u = G.getEdge(i).first  + 1;
        int v = G.getEdge(i).second + 1;
        fprintf(fp, "e %d %d\n", u, v);
    }
    fprintf(fp, "l %d\n", l);
    if(s > 0 && t > 0) fprintf(fp, "t %d %d\n", s, t);
    fclose(fp);
}